# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Wraps Julia such that its low-level functions can be used by PyRates to create and simulate a compute graph.
"""

# pyrates internal _imports
import sys

from ..base import BaseBackend
from ..computegraph import ComputeVar
from .julia_funcs import julia_funcs

# external _imports
from typing import Optional, Dict, List, Callable, Iterable, Union, Tuple
import numpy as np

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


# backend classes
#################


class JuliaBackend(BaseBackend):

    def __init__(self,
                 ops: Optional[Dict[str, str]] = None,
                 imports: Optional[List[str]] = None,
                 **kwargs
                 ) -> None:
        """Instantiates tensorflow backend, i.e. a tensorflow graph.
        """

        # add user-provided operations to function dict
        julia_ops = julia_funcs.copy()
        if ops:
            julia_ops.update(ops)

        # set default float precision to float64
        kwargs["float_precision"] = "float64"

        # call parent method
        super().__init__(ops=julia_ops, imports=imports, file_ending='.jl', start_idx=1, **kwargs)

        # define fortran-specific imports
        self._imports.pop(0)
        self._imports.append("using LinearAlgebra")

        # set up pyjulia
        from julia.api import Julia
        jl = Julia(runtime=kwargs.pop('julia_path'), compiled_modules=False)
        from julia import Main
        self._jl = Main
        self._no_vectorization = ["*(", "interp("]

    def get_var(self, v: ComputeVar):
        v = super().get_var(v)
        dtype = v.dtype.name
        s = sum(v.shape)
        if s > 0:
            return v
        if 'float' in dtype:
            return float(v)
        if 'complex' in dtype:
            return complex(np.real(v), np.imag(v))
        return int(v)

    def add_var_update(self, lhs: ComputeVar, rhs: str, lhs_idx: Optional[str] = None, rhs_shape: Optional[tuple] = ()):

        super().add_var_update(lhs=lhs, rhs=rhs, lhs_idx=lhs_idx, rhs_shape=rhs_shape)
        if rhs_shape or lhs_idx:
            line = self.code.pop()
            lhs, rhs = line.split(' = ')
            if not any([rhs[:len(expr)] == expr for expr in self._no_vectorization]):
                rhs = f"@. {rhs}"
            self.add_code_line(f"{lhs} = {rhs}")

    def create_index_str(self, idx: Union[str, int, tuple], separator: str = ',', apply: bool = True,
                         **kwargs) -> Tuple[str, dict]:

        if not apply:
            self._start_idx = 0
            idx, idx_dict = super().create_index_str(idx, separator, apply, **kwargs)
            self._start_idx = 1
            return idx, idx_dict
        else:
            return super().create_index_str(idx, separator, apply, **kwargs)

    def generate_func_tail(self, rhs_var: str = None):

        self.add_code_line(f"return {rhs_var}")
        self.remove_indent()
        self.add_code_line("end")

    def generate_func(self, func_name: str, to_file: bool = True, **kwargs):

        # generate the current function string via the code generator
        if kwargs.pop('julia_diffeq', False):
            self.add_linebreak()
            self.add_code_line(f"function {func_name}_julia(dy, y, p, t)")
            self.add_indent()
            self.add_code_line(f"return {func_name}(t, y, dy, p...)")
            self.remove_indent()
            self.add_code_line("end")
        func_str = self.generate()

        if to_file:

            # save rhs function to file
            file = f'{self._fdir}/{self._fname}{self._fend}' if self._fdir else f"{self._fname}{self._fend}"
            with open(file, 'w') as f:
                f.writelines(func_str)
                f.close()

            # import function from file
            rhs_eval = self._jl.include(file)

        else:

            # just execute the function string, without writing it to file
            rhs_eval = self._jl.eval(func_str)

        # apply function decorator
        decorator = kwargs.pop('decorator', None)
        if decorator:
            decorator_kwargs = kwargs.pop('decorator_kwargs', dict())
            rhs_eval = decorator(rhs_eval, **decorator_kwargs)

        return rhs_eval

    def _solve(self, solver: str, func: Callable, args: tuple, T: float, dt: float, dts: float, y0: np.ndarray,
               t0: np.ndarray, times: np.ndarray, **kwargs) -> np.ndarray:

        # perform integration via scipy solver (mostly Runge-Kutta methods)
        if solver == 'euler':

            # solve ivp via forward euler method (fixed integration step-size)
            results = self._solve_euler(func, args, T, dt, dts, y0, t0)

        elif solver == 'scipy':

            # solve ivp via scipy methods (solvers of various orders with adaptive step-size)
            from scipy.integrate import solve_ivp
            kwargs['t_eval'] = times

            # call scipy solver
            results = solve_ivp(fun=func, t_span=(t0, T), y0=y0, first_step=dt, args=args, **kwargs)
            results = results['y'].T

        else:

            # solve ivp via DifferentialEquations.jl solver
            self._jl.eval('using DifferentialEquations')
            model = self._jl.ODEProblem(func, y0, [0.0, T], args[1:])
            method = kwargs.pop('method', 'Tsit5')
            if hasattr(self._jl, method):
                method = getattr(self._jl, method)
            else:
                method = self._jl.Tsit5
            results = self._jl.solve(model, method(), saveat=times, reltol=1e-6, abstol=1e-6)
            results = np.asarray(results).T

        return results

    def _add_func_call(self, name: str, args: Iterable):
        self.add_code_line(f"function {name}({','.join(args)})")

    def _process_idx(self, idx: Union[Tuple[int, int], int, str, ComputeVar], **kwargs) -> str:

        if type(idx) is str and ':' in idx:
            idx0, idx1 = idx.split(':')
            self._start_idx = 0
            idx0 = int(self._process_idx(idx0))
            idx1 = int(self._process_idx(idx1))
            self._start_idx = 1
            return self._process_idx((idx0, idx1))
        return super()._process_idx(idx=idx, **kwargs)

    @staticmethod
    def expr_to_str(expr: str, args: tuple):

        # replace power operator
        func = '**'
        while func in expr:
            expr = expr.replace(func, '^')

        return expr



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""This module provides the backend class that should be used to set up any backend in pyrates.
"""

# external _imports
from typing import Any, Callable, Union, Iterable, Optional
from networkx import MultiDiGraph
from sympy import Symbol, Expr, Function
import numpy as np

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


#########################
# compute graph classes #
#########################


# numpy-based node class
class ComputeNode:
    """Base class for adding variables to the compute graph. Creates a numpy array with additional attributes
    for variable identification/retrieval from graph. Should be used as parent class for custom variable classes.

    Parameters
    ----------
    name
        Full name of the variable in the original NetworkGraph (including the node and operator it belongs to).
    dtype
        Data-type of the variable. For valid data-types, check the documentation of the backend in use.
    shape
        Shape of the variable.
    """

    __slots__ = ["name", "symbol", "dtype", "shape", "_value"]

    def __init__(self, name: str, symbol: Union[Symbol, Expr, Function], dtype: Optional[str] = None,
                 shape: Optional[tuple] = None, def_shape: Optional[tuple] = None):
        """Instantiates a basic node of a ComputeGraph instance.
        """

        self.name = name
        self.symbol = symbol
        self.shape = self._get_shape(shape, def_shape)
        self._value = np.zeros(self.shape)
        self.dtype = dtype
        self._set_dtype()

    def reshape(self, shape: tuple, **kwargs):

        self._value = self.value.reshape(shape, **kwargs)
        self.shape = shape
        return self

    def squeeze(self, axis=None):
        self._value = self.value.squeeze(axis=axis)
        self.shape = self._value.shape
        return self

    def set_value(self, v: Union[float, np.ndarray]):
        self._value = np.asarray(v, dtype=self.dtype)
        self.shape = tuple(v.shape)

    @property
    def value(self):
        """Returns current value of BaseVar.
        """
        return self._value

    @property
    def is_constant(self):
        raise NotImplementedError("This method has to be defined by each child class.")

    @property
    def is_float(self):
        return "float" in self.dtype

    @property
    def is_complex(self):
        return "complex" in self.dtype

    def _is_equal_to(self, v):
        for attr in self.__slots__:
            if not hasattr(v, attr) or getattr(v, attr) != getattr(self, attr):
                return False
        return True

    def _get_value(self, value: Optional[Union[list, np.ndarray]] = None, dtype: Optional[str] = None,
                   shape: Optional[tuple] = None):
        """Defines initial value of variable.
        """

        # case I: create new array from shape and dtype
        if value is None:
            return np.zeros(shape=shape, dtype=dtype)

        # case II: transform values into an array
        if not hasattr(value, 'shape'):
            if type(value) is list:
                return self._get_value(value=np.asarray(value, dtype=dtype), dtype=dtype, shape=shape)
            return np.zeros(shape=shape, dtype=dtype) + value

        # case III: match given shape with the shape of the given value array
        if shape is not None:
            value = np.asarray(value, dtype=dtype)
            if value.shape == shape:
                return value
            if sum(shape) < sum(value.shape):
                return value.squeeze()
            idx = ",".join("None" if s == 1 else ":" for s in shape)
            return eval(f'value[{idx}]')

        # case IV: just ensure the correct data type of the value array
        return np.asarray(value, dtype=dtype)

    def _set_dtype(self):
        if not self.dtype:
            if 'float' in str(self.value.dtype):
                self.dtype = 'float'
            elif 'complex' in str(self.value.dtype):
                self.dtype = 'complex'
            else:
                self.dtype = 'int'

    @staticmethod
    def _get_shape(s: Union[tuple, None], s_def: tuple):
        if s is None or sum(s) < 2:
            return s_def
        return s

    def __deepcopy__(self, memodict: dict):
        node = ComputeNode(name=self.name, symbol=self.symbol, dtype=self.dtype, shape=self.shape)
        node._value = np.zeros_like(node._value) + node._value
        return node

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(str(self))


class ComputeVar(ComputeNode):
    """Class for variables and vector-valued constants in the ComputeGraph.
    """

    __slots__ = ComputeNode.__slots__ + ["vtype"]

    def __init__(self, name: str, symbol: Union[Symbol, Expr, Function], vtype: str, dtype: Optional[str] = None,
                 shape: Optional[str] = None, value: Optional[Union[list, np.ndarray]] = None,
                 def_shape: Optional[tuple] = None):

        # set attributes
        super().__init__(name=name, symbol=symbol, dtype=dtype, shape=shape, def_shape=def_shape)
        self.vtype = vtype

        # adjust variable value
        self.set_value(self._get_value(value=value, shape=self.shape, dtype=self.dtype))

    @property
    def is_constant(self):
        return self.vtype == 'constant'


class ComputeOp(ComputeNode):
    """Class for ComputeGraph nodes that represent mathematical operations.
    """

    __slots__ = ComputeNode.__slots__ + ["func", "expr"]

    def __init__(self, name: str, symbol: Union[Symbol, Expr, Function], func: Callable, expr: Expr,
                 dtype: Optional[str] = None, shape: Optional[str] = None):

        # set attributes
        super().__init__(name=name, symbol=symbol, dtype=dtype, shape=shape)
        self.func = func
        self.expr = expr

    @property
    def is_constant(self):
        return False


# networkx-based graph class
class ComputeGraph(MultiDiGraph):
    """Creates a compute graph where nodes are all constants and variables of the network and edges are the mathematical
    operations linking those variables/constants together to form equations.
    """

    def __init__(self, backend: str, **kwargs):

        super().__init__()

        # choose a backend
        if backend == 'tensorflow':
            from pyrates.backend.tensorflow import TensorflowBackend
            backend = TensorflowBackend
        elif backend == 'torch':
            from pyrates.backend.torch import TorchBackend
            backend = TorchBackend
        elif backend == 'fortran':
            from pyrates.backend.fortran import FortranBackend
            backend = FortranBackend
        elif backend == 'julia':
            from pyrates.backend.julia import JuliaBackend
            backend = JuliaBackend
        else:
            from pyrates.backend.base import BaseBackend
            backend = BaseBackend

        # backend-related attributes
        self.backend = backend(**kwargs)
        self.var_updates = {'DEs': dict(), 'non-DEs': dict()}
        self._eq_nodes = []
        self._state_var_indices = dict()

    @property
    def state_vars(self):
        return list(self.var_updates['DEs'].keys())

    def add_var(self, label: str, value: Any, vtype: str, **kwargs):

        unique_label = self._generate_unique_label(label)
        var = ComputeVar(name=unique_label, symbol=Symbol(unique_label), value=value, vtype=vtype, **kwargs)
        super().add_node(unique_label, node=var)
        return unique_label, self.nodes[unique_label]['node']

    def add_op(self, inputs: Union[list, tuple], label: str, expr: Expr, func: Callable, **kwargs):

        # add target node that contains result of operation
        unique_label = self._generate_unique_label(label)
        op = ComputeOp(name=unique_label, symbol=Symbol(unique_label), func=func, expr=expr, **kwargs)
        super().add_node(unique_label, node=op)

        # add edges from source nodes to target node
        for i, v in enumerate(inputs):
            super().add_edge(v, unique_label, key=i)

        return unique_label, self.nodes[unique_label]['node']

    def add_var_update(self, var: str, update: str, differential_equation: bool = False):

        # store mapping between left-hand side variable and right-hand side update
        if differential_equation:
            self.var_updates['DEs'][var] = update
        else:
            self.var_updates['non-DEs'][var] = update

        # remember var and update node to ensure that they are not pruned during compilation
        self._eq_nodes.extend([var, update])

    def get_var(self, var: str, from_backend: bool = False):
        v = self.nodes[var]['node']
        if from_backend:
            return self.backend.get_var(v)
        return v

    def get_op(self, op: str, **kwargs) -> dict:
        return self.backend.get_op(op, **kwargs)

    def eval_graph(self):

        for n in self.var_updates['non-DEs'].values():
            self.eval_subgraph(n)
        return self.eval_nodes(self.var_updates['DEs'].values())

    def eval_nodes(self, nodes: Iterable):

        return [self.eval_node(n) for n in nodes]

    def eval_node(self, n):

        inputs = [self.eval_node(inp) for inp in self.predecessors(n)]
        node = self.get_var(n)
        if isinstance(node, ComputeOp):
            return node.func(*tuple(inputs))
        return node.value

    def eval_subgraph(self, n):

        inputs = []
        input_nodes = [node for node in self.predecessors(n)]
        for inp in input_nodes:
            inputs.append(self.eval_subgraph(inp))
            self.remove_node(inp)

        node = self.get_var(n)
        if inputs:
            node.set_value(node.func(*tuple(inputs)))

        return node.value

    def remove_subgraph(self, n):

        for inp in self.predecessors(n):
            self.remove_subgraph(inp)
        self.remove_node(n)

    def compile(self):

        # evaluate constant-based operations
        out_nodes = [node for node, out_degree in self.out_degree if out_degree == 0]
        for node in out_nodes:

            # process inputs of node
            for inp in self.predecessors(node):
                if self.get_var(inp).is_constant:
                    self.eval_subgraph(inp)

            # evaluate node if all its inputs are constants
            if all([self.get_var(inp).is_constant for inp in self.predecessors(node)]):
                self.eval_subgraph(node)

        # remove unconnected nodes and constants from graph
        self._prune()

        return self

    def to_func(self, func_name: str, to_file: bool = True, dt_adapt: bool = True, **kwargs):

        # finalize compute graph
        self.compile()

        # create state variable vector and state variable update vector
        ###############################################################

        variables = []
        idx = 0
        #dtype = 'float'
        for var, update in self.var_updates['DEs'].items():

            # extract left-hand side nodes from graph
            lhs, rhs = self._process_var_update(var, update)
            variables.append(lhs.value)
            #if 'complex' in lhs.dtype:
            #    dtype = 'complex'

            # store information of the original, non-vectorized state variable
            vshape = sum(lhs.shape)
            if vshape >= 1:
                self._state_var_indices[var] = (idx, idx+vshape)
                idx += vshape
            else:
                self._state_var_indices[var] = idx
                idx += 1

        # add collected state variables to the backend
        try:
            state_vec = np.concatenate(variables, axis=0)
        except ValueError:
            state_vec = np.asarray(variables)
        dtype = 'complex' if 'complex' in state_vec.dtype.name else 'float'
        state_var_key, y = self.add_var(label='y', vtype='state_var', value=state_vec, dtype=dtype)
        rhs_var_key, dy = self.add_var(label='dy', vtype='state_var', value=np.zeros_like(state_vec), dtype=dtype)
        try:
            t = self.get_var('t')
        except KeyError:
            _, t = self.add_var(label='t', vtype='state_var', value=0.0 if dt_adapt else 0,
                                dtype='float' if dt_adapt else 'int', shape=())
        self.backend.register_vars([t, y])

        # create a string containing all computations and variable updates represented by the compute graph
        func_args, code_gen = self._to_str()
        func_body = code_gen.generate()
        code_gen.code.clear()

        # generate function head
        func_args = code_gen.generate_func_head(func_name=func_name, state_var=state_var_key,
                                                func_args=[self.get_var(arg) for arg in func_args])

        # add lines from function body after function head
        code_gen.add_linebreak()
        code_gen.add_code_line(func_body)
        code_gen.add_linebreak()

        # generate function tail
        code_gen.generate_func_tail(rhs_var=rhs_var_key)

        # generate the function (and write to file, optionally)
        func = code_gen.generate_func(func_name=func_name, to_file=to_file, func_args=func_args[3:],
                                      state_vars=self.state_vars, **kwargs)

        # OPTIONAL: write function arguments (state vectors and constants) to file
        c_fn = kwargs.pop('constants_file_name', None)
        if c_fn:
            arg_dict = {arg: self.get_var(arg).value for arg in func_args}
            fn = f'{self.backend._fdir}/{c_fn}' if self.backend._fdir else c_fn
            np.savez(fn, **arg_dict)

        return func, tuple([self.get_var(arg, from_backend=True) for arg in func_args])

    def run(self, func: Callable, func_args: tuple, T: float, dt: float, dts: Optional[float] = None,
            outputs: Optional[dict] = None, **kwargs):

        # pre-process outputs
        if outputs is None:
            outputs = {key: key for key in self.state_vars}
        for key in outputs.copy():
            var = outputs.pop(key)
            outputs[key] = self._state_var_indices[var]

        # handle other arguments
        if dts is None:
            dts = dt
        solver = kwargs.pop('solver', 'euler')

        # call backend method
        return self.backend.run(func=func, func_args=func_args, T=T, dt=dt, dts=dts, outputs=outputs, solver=solver,
                                **kwargs)

    def clear(self) -> None:
        """Deletes build directory and removes all compute graph nodes
        """

        # delete network nodes and variables from the compute graph
        for n in list(self.nodes.keys()):
            self.remove_subgraph(n)
        self.var_updates.clear()
        self._state_var_indices.clear()
        self._eq_nodes.clear()

        # clear code generator
        self.backend.clear()

    def _to_str(self):

        # preparations
        code_gen = self.backend

        # extract state variable from state vector
        rhs_indices_str = []
        for var in self.state_vars:

            # extract index of variable in state vector
            idx = (self._state_var_indices[var],)

            # extract state variable from state vector
            rhs_idx, _ = code_gen.create_index_str(idx)
            code_gen.add_var_update(lhs=self.get_var(var), rhs=f"y{rhs_idx}")
            rhs_indices_str.append(idx)

        code_gen.add_linebreak()

        # get equation string and argument list for each non-DE node at the end of the compute graph hierarchy
        func_args2, delete_args1 = self._generate_update_equations(differential_equations=False)
        code_gen.add_linebreak()

        # get equation string and argument list for each DE node at the end of the compute graph hierarchy
        func_args1, delete_args2 = self._generate_update_equations(differential_equations=True, indices=rhs_indices_str)

        # remove unnecessary function arguments
        func_args = func_args1 + func_args2
        for arg in delete_args1 + delete_args2:
            while arg in func_args:
                func_args.pop(func_args.index(arg))

        return func_args, code_gen

    def _generate_update_equations(self, differential_equations: bool, indices: list = None) -> tuple:

        code_gen = self.backend

        # extract relevant compute graph nodes and bring them into the correct order
        nodes = self.var_updates['DEs' if differential_equations else 'non-DEs']
        nodes, updates, def_vars, undef_vars = self._sort_var_updates(nodes=nodes,
                                                                      differential_equations=differential_equations)

        # collect right-hand side expression and all input variables to these expressions
        func_args, expressions, var_names, rhs_shapes, lhs_indices = undef_vars, [], [], [], []
        for node, update in zip(nodes, updates):

            # collect shape of the right-hand side variable
            v = self.get_var(update)
            try:
                v_eval = self.eval_node(update)
                v.set_value(v_eval)
            except IndexError:
                pass
            rhs_shapes.append(v.shape)

            # collect expression and variables of right-hand side of equation
            expr_args, expr = self._node_to_expr(update)
            func_args.extend(expr_args)
            expr_str, expr_args, _, _ = self._expr_to_str(expr, apply=True)
            func_args.extend(expr_args)
            expressions.append(expr_str)

            # process left-hand side of equation
            var = self.get_var(node)
            if isinstance(var, ComputeOp):

                # process indexing of left-hand side variable
                idx_args, lhs = self._node_to_expr(node)
                if lhs.args[0].name not in def_vars:
                    idx_args.append(lhs.args[0].name)
                func_args.extend(idx_args)
                _, idx_args, lhs_var, idx = self._expr_to_str(lhs, apply=False)
                func_args.extend(idx_args)

            else:

                # process normal update of left-hand side variable
                lhs_var = var.name
                idx = None

            var_names.append(lhs_var)
            lhs_indices.append(idx)

        # add the left-hand side assignments of the collected right-hand side expressions to the code generator
        if differential_equations:

            # differential equation (DE) update
            if not indices:
                raise ValueError('State variables need to be stored in a single state vector, for which the indices '
                                 'have to be passed to this method.')

            # DE updates stored in a state-vector
            dy = self.get_var("dy")
            for idx, expr, shape in zip(indices, expressions, rhs_shapes):
                code_gen.add_var_update(lhs=dy, rhs=expr, lhs_idx=idx, rhs_shape=shape)

            # add rhs var to function arguments
            func_args = ['dy'] + func_args

        else:

            # non-differential equation update
            if indices:
                raise ValueError('Indices to non-state variables should be defined in the respective equations, not'
                                 'be passed to this method.')
            indices = lhs_indices

            # non-DE update stored in a single variable
            for target_var, expr, idx, shape in zip(var_names, expressions, indices, rhs_shapes):
                try:
                    idx = self.get_var(idx)
                except (KeyError, AttributeError):
                    pass
                code_gen.add_var_update(lhs=self.get_var(target_var), rhs=expr, lhs_idx=idx, rhs_shape=shape)

        return func_args, def_vars

    def _node_to_expr(self, n: str, **kwargs) -> tuple:

        expr_args = []
        node = self.get_var(n)

        # case I: node is a mathematical operation and its inputs need to be treated
        try:

            # process node inputs
            expr_info = {self.get_var(inp).symbol: self._node_to_expr(inp, **kwargs) for inp in self.predecessors(n)}

            # replace old inputs with its processed versions
            expr = node.expr
            for expr_old, (args, expr_new) in expr_info.items():
                if expr_old != expr_new:
                    expr = expr.replace(expr_old, expr_new)
                expr_args.extend(args)

            # replace generic function calls with the backend-specific function calls
            try:
                expr_old = expr.func.__name__
                func_info = self.get_op(expr_old, shape=node.shape)
                expr = expr.replace(Function(expr_old), Function(func_info['call']))
            except (AttributeError, KeyError):
                pass

        # case II: node is a simple variable or constant
        except AttributeError:

            # add constants to the expression arguments list
            if node.is_constant:
                expr_args.append(n)
                expr = node.symbol
            elif 'dummy_constant' in node.name:
                expr = Symbol(str(float(node.value)))
            else:
                expr = node.symbol

        return expr_args, expr

    def _expr_to_str(self, expr: Any, expr_str: str = None, apply: bool = True, **kwargs) -> tuple:

        # preparations
        ###############

        # initializations
        index_args = []
        func = ""
        var = str(expr.args[0]) if expr.args else ""
        idx = ""

        # ensure expression string exists
        if not expr_str:
            expr_str = str(expr)

            # transform expression arguments into strings
            for arg in expr.args:
                expr_part, args, _, _ = self._expr_to_str(arg, **kwargs)
                expr_str = expr_str.replace(str(arg), expr_part)
                index_args.extend(args)

        # process indexing operations
        #############################

        if 'index_1d(' in expr_str:

            # replace `index` calls with brackets-based indexing
            idx = self._get_var_idx(idx=(expr.args[1],), args=index_args, apply=apply, **kwargs)
            func = 'index_1d'

        elif 'index_2d(' in expr_str:

            # replace `2d_index` calls with brackets-based indexing
            idx = self._get_var_idx(idx=(expr.args[1], expr.args[2]), args=index_args, apply=apply, **kwargs)
            func = 'index_2d'

        elif 'index_range(' in expr_str:

            # replace `range_index` calls with brackets-based indexing
            idx = self._get_var_idx(idx=((expr.args[1], expr.args[2]),), args=index_args, apply=apply, **kwargs)
            func = 'index_range'

        elif 'index_axis(' in expr_str:

            # replace `axis_index` calls with brackets-based indexing
            if len(expr.args) < 2:
                idx = self._get_var_idx(idx=(':',), args=index_args, apply=apply, **kwargs)
            else:
                idx = self._get_var_idx(args=index_args, apply=apply,
                                        idx=tuple([':' for _ in range(expr.args[2])] + [f"{expr.args[1]}"]), **kwargs)
            func = "index_axis"

        # either apply the above indexing calls or return them
        if func and apply:
            replacement = self.backend.finalize_idx_str(var=self.get_var(var), idx=idx)
            expr_str = self._process_func_call(expr=expr_str, func=func, replacement=replacement)

        # handle other function calls
        #############################

        if 'identity(' in expr_str:

            # replace `no_op` calls with first argument to the function call
            expr_str = self._process_func_call(expr=expr_str, func='identity', replacement=var)

        # backend-specific function call adjustments
        expr_str = self.backend.expr_to_str(expr_str, expr.args)

        return expr_str, index_args, var, idx

    def _process_var_update(self, var: str, update: str) -> tuple:

        # extract nodes
        lhs = self.get_var(var)
        rhs = self.eval_node(update)

        # extract common shape
        if lhs.shape == rhs.shape:
            return lhs, rhs
        try:
            rhs = rhs.reshape(lhs.shape)
            return lhs, rhs
        except ValueError:
            raise ValueError(
                f"Shapes of state variable {var} and its right-hand side update {self.get_var(update).expr} do not"
                " match.")

    def _sort_var_updates(self, nodes: dict, differential_equations: bool = True) -> tuple:

        # case I: for differential equations, do not perform any sorting
        if differential_equations:
            return list(nodes.keys()), list(nodes.values()), [], []

        # case II: for non-differential equations, sort them according to their graph connections
        #########################################################################################

        # step 1: ensure lhs-indexing operations are considered as well
        node_names, node_keys = [], []
        for node in nodes:
            n = self.get_var(node)
            if type(n) is ComputeVar:
                node_names.append(node)
            else:
                node_names.append(list(self._get_inputs(node))[0])
            node_keys.append(node)

        keys, values, defined_vars, undefined_vars = [], [], [], []
        n_nodes = len(nodes)
        while nodes:

            for node, update in nodes.copy().items():

                # go through node inputs and check whether other it depends on other equations to be evaluated first
                dependent, inp = False, ""
                for inp in self._get_inputs(update):
                    if inp in node_names:
                        idx = node_names.index(inp)
                        if node_keys[idx] != node:
                            dependent = True
                            break

                # decide whether this equation can be evaluated now
                if dependent:
                    continue
                else:
                    idx = node_keys.index(node)
                    n = node_names.pop(idx)
                    node_keys.pop(idx)
                    nodes.pop(node)
                    keys.append(node)
                    values.append(update)
                    if isinstance(self.get_var(node), ComputeVar) and n not in undefined_vars:
                        defined_vars.append(n)
                    elif n not in undefined_vars:
                        undefined_vars.append(n)

            # check whether the algorithm is stuck
            n_nodes_new = len(nodes)
            if n_nodes_new == n_nodes:
                break
            else:
                n_nodes = n_nodes_new

        # add mutually depended nodes
        if nodes:
            node_keys = list(nodes.keys())
            keys.extend(node_keys)
            values.extend(list(nodes.values()))
            for n in node_names:
                if n not in undefined_vars:
                    undefined_vars.append(n)

        return keys, values, defined_vars, undefined_vars

    def _get_inputs(self, n: str):

        inputs = []
        for inp in self.predecessors(n):
            inputs.extend([inp] if isinstance(self.get_var(inp), ComputeVar) else self._get_inputs(inp))
        return inputs

    def _get_var_idx(self, idx: tuple, args: list, apply: bool = True, **kwargs):

        # collect indexing variables where necessary
        new_idx = []
        for idx_tmp in idx:
            try:
                v = self.get_var(idx_tmp.name if type(idx_tmp) is Symbol else idx_tmp)
                new_idx.append(v)
            except KeyError:
                new_idx.append(idx_tmp)

        # turn index into a backend-specific string
        idx_str, new_vars = self.backend.create_index_str(tuple(new_idx), apply=apply, **kwargs)

        # add new variables to graph and index arguments
        for key, v in new_vars.items():
            vlabel, _ = self.add_var(label='index', value=v, vtype='constant')
            idx_str = idx_str.replace(key, vlabel)
            args.append(vlabel)

        return idx_str

    def _prune(self):

        # remove all subgraphs that contain constants only
        for n in [node for node, out_degree in self.out_degree if out_degree == 0]:
            if self.get_var(n).is_constant and n not in self._eq_nodes:
                self.remove_subgraph(n)

        # remove all unconnected nodes
        for n in [node for node, out_degree in self.out_degree if out_degree == 0]:
            if self.in_degree(n) == 0 and n not in self._eq_nodes:
                self.remove_node(n)

    def _generate_unique_label(self, label: str):

        if label in self.nodes:
            label_split = label.split('_')
            try:
                new_label = "_".join(label_split[:-1] + [f"{int(label_split[-1])+1}"])
            except ValueError:
                new_label = f"{label}_0"
            return self._generate_unique_label(new_label)
        else:
            return label

    @staticmethod
    def _process_func_call(expr: str, func: str, replacement: str):

        # identify start and end of the function call
        start = expr.find(f"{func}(")
        end = expr[start:].find(')') + 1

        # replace part in expression string
        return expr.replace(expr[start:start + end], replacement)



"""Test suit for saving PyRates objects to or loading them from file."""

__author__ = "Daniel Rose"
__status__ = "Development"

import filecmp
import os
import pathlib

import pytest


def setup_module():
    print("\n")
    print("=======================")
    print("| Test Suite File I/O |")
    print("=======================")


class FileCompareError(Exception):
    pass


def compare_files(filename1, filename2):
    """Helper function to compare files"""

    if os.path.getsize(filename1) == os.path.getsize(filename2):
        if filecmp.cmp(filename1, filename2, shallow=False):
            pass
        else:
            raise FileCompareError("Files have different content.")
    else:
        raise FileCompareError("Files are not of the same size.")


def get_parent_directory():
    """Helper function to get path of current file"""
    return pathlib.Path(__file__).parent.absolute()


def test_yaml_template():
    path = "model_templates.neural_mass_models.jansenrit.JRC"
    from pyrates import clear_frontend_caches
    from pyrates import save, CircuitTemplate
    clear_frontend_caches()

    template = CircuitTemplate.from_yaml(path)

    # configure filenames
    out_file = os.path.join(get_parent_directory(), "output", "jansen_rit_template.yaml")
    test_file = os.path.join(get_parent_directory(), "resources", "jansen_rit_template.yaml")

    template.to_yaml(out_file)
    # to update the reference dump, uncomment the following
    # save(template, test_file, filetype='yaml', template_name='jrc1')

    compare_files(out_file, test_file)

    path = f"{out_file.split('.')[0]}/JRC"
    template = CircuitTemplate.from_yaml(path)
    assert template



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Contains fortran function definitions that may be used for PyRates model equations.
"""

from typing import Union
import numpy as np

# function definitions
######################


# sigmoid function
sigmoid = lambda x: 1./(1. + np.exp(-x))


# function for calculating the logistic function of an nd-array
def get_sigmoid_def(idx: int, out_shape: Union[tuple, str] = '', dtype: str = 'real') -> tuple:
    isarray = len(out_shape) > 0
    fname = f"fsigmoid_{idx}"
    sigmoid_n = f"s = size(x)\ndo n=1,s\n  {fname}(n) = 1 / (1 + exp(-x(n)))\nend do"
    sigmoid_0 = f"{fname} = 1 / (1 + exp(-x))"

    func = f"""
function {fname}(x)

implicit none

{dtype} :: {fname}{out_shape if isarray else ''}
{dtype} :: x{out_shape if isarray else ''}
{"integer :: n, s" if isarray else ""}

{sigmoid_n if isarray else sigmoid_0}

end function {fname}
    """
    return fname, func


# wrapper function for interpolating a 1d-array
def get_interp_def(idx: int, out_shape: Union[tuple, str] = '', dtype: str = 'real') -> tuple:
    fname = f"finterp_{idx}"
    func = f"""
function {fname}(x_new,x,y)

implicit none 

{dtype} :: {fname}
{dtype} :: x(:), y(:), x_new
{dtype} :: x_inc
integer :: n, s

s = size(x) 

if (x_new < x(1)) then
  {fname} = y(1)
else if (x_new > x(s)) then
  {fname} = y(s)
else
  do n = 1, s 
    if (x(n) > x_new) exit 
  end do
  if (n == 1) then 
    {fname} = y(1)
  else if (n == s+1) then
    {fname} = y(s)
  else
    x_inc = (x_new - x(n-1)) / (x(n) - x(n-1))
    {fname} = y(n) + x_inc*(y(n) - y(n-1))
  end if 
end if 

end function {fname}
"""
    return fname, func


# dictionary for backend import
###############################

fortran_funcs = {
    'max': {'call': 'max', 'func': np.maximum},
    'min': {'call': 'min', 'func': np.minimum},
    'round': {'call': 'nint', 'func': np.round},
    'sum': {'call': 'sum', 'func': np.sum},
    'matmul': {'call': 'matmul', 'func': np.dot},
    'matvec': {'call': 'matmul', 'func': np.dot},
    'roll': {'call': 'cshift', 'func': np.roll},
    'tanh': {'call': 'tanh', 'func': np.tanh},
    'sinh': {'call': 'sinh', 'func': np.sinh},
    'cosh': {'call': 'cosh', 'func': np.cosh},
    'arctan': {'call': 'atan', 'func': np.arctan},
    'arcsin': {'call': 'asin', 'func': np.arcsin},
    'arccos': {'call': 'acos', 'func': np.arccos},
    'sin': {'call': 'sin', 'func': np.sin},
    'cos': {'call': 'cos', 'func': np.cos},
    'tan': {'call': 'tan', 'func': np.tan},
    'exp': {'call': 'exp', 'func': np.exp},
    'sigmoid': {'call': get_sigmoid_def, 'func': sigmoid},
    'interp': {'call': get_interp_def, 'func': np.interp},
    'real': {'call': 'realpart', 'func': np.real},
    'imag': {'call': 'imagpart', 'func': np.imag},
    'conj': {'call': 'conjg', 'func': np.conjugate},
    'absv': {'call': 'abs', 'func': np.abs},
}



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Contains wrapper classes for different backends that are needed by the parser module.

A new backend needs to implement the following methods:
- __init__
- run
- add_var
- add_op
- add_layer

Currently supported backends:
- Numpy: BaseBackend.
- Tensorflow: TensorflowBackend.
- Fortran: FortranBackend (experimental).

"""

# pyrates internal _imports
from ..computegraph import ComputeVar
from .base_funcs import base_funcs

# external _imports
from typing import Optional, Dict, List, Union, Tuple, Callable, Iterable
import numpy as np
import os, sys
from shutil import rmtree


# Helper Functions and Classes
##############################


class CodeGen:
    """Generates python code. Can add code lines, line-breaks, indents and remove indents.
    """

    def __init__(self):
        self.code = []
        self.lvl = 0

    def generate(self):
        """Generates a single code string from its history of code additions.
        """
        return '\n'.join(self.code)

    def add_code_line(self, code_str):
        """Add code line string to code.
        """
        code_str = code_str.split('\n')
        for code in code_str:
            self.code.append("\t" * self.lvl + code)

    def add_linebreak(self):
        """Add a line-break to the code.
        """
        self.code.append("")

    def add_indent(self):
        """Add an indent to the code.
        """
        self.lvl += 1

    def remove_indent(self):
        """Remove an indent to the code.
        """
        if self.lvl == 0:
            raise(SyntaxError("Error in generation of network function file: A net negative indentation was requested.")
                  )
        self.lvl -= 1

    def clear(self):
        """Deletes all code lines from the memory of the generator.
        """
        self.code.clear()


#######################################
# classes for backend functionalities #
#######################################


class BaseBackend(CodeGen):
    """Default backend class. Transforms all network equations into their numpy equivalents. Based on a Python code
    generator.
    """

    def __init__(self,
                 ops: Optional[Dict[str, str]] = None,
                 imports: Optional[List[str]] = None,
                 **kwargs
                 ) -> None:
        """Instantiates the standard, numpy-based backend, i.e. a compute graph with numpy operations.

        Parameters
        ----------
        ops
        imports
        kwargs
        """

        # call to super method (initializes code generator)
        super().__init__()

        # definition of usable math operations
        self._funcs = base_funcs.copy()
        if ops:
            self._funcs.update(ops)
        self._helper_funcs = []

        # definition of extrinsic function _imports
        self._imports = ["from numpy import pi, sqrt"]
        if imports:
            for imp in imports:
                if imp not in self._imports:
                    self._imports.append(imp)

        # private attributes
        self._float_precision = kwargs.pop('float_precision', 'float32')
        self._int_precision = kwargs.pop('int_precision', 'int32')
        self._complex_precision = kwargs.pop('complex_precision', 'complex64')
        self._idx_left = kwargs.pop('idx_left', '[')
        self._idx_right = kwargs.pop('idx_right', ']')
        self._start_idx = kwargs.pop('start_idx', 0)
        self._no_funcs = ["identity", "index_1d", "index_2d", "index_range", "index_axis"]

        # file-creation-related attributes
        fdir, fname = self.get_fname(kwargs.pop('file_name', 'pyrates_run'))
        if fdir:
            sys.path.append(fdir)
        else:
            sys.path.append(os.getcwd())
        self._fdir = fdir
        self._fname = fname
        self._fend = kwargs.pop('file_ending', '.py')

    def get_var(self, v: ComputeVar):
        if v.is_float:
            dtype = self._float_precision
        elif v.is_complex:
            dtype = self._complex_precision
        else:
            dtype = self._int_precision
        return np.asarray(v.value, dtype=dtype)

    def get_op(self, name: str, **kwargs) -> dict:

        # retrieve function information from backend definitions
        func_info = self._get_func_info(name, **kwargs)
        func_name = func_info['call']

        # add extrinsic function _imports if necessary
        if 'imports' in func_info:
            for imp in func_info['imports']:
                *in_path, in_func = imp.split('.')
                import_line = f"from {'.'.join(in_path)} import {in_func}"
                if import_line not in self._imports:
                    self._imports.append(import_line)

        if 'def' in func_info:

            # extract the provided function definition
            func_str = func_info['def']

            # remember the function definition string for file creation
            if func_str not in self._helper_funcs and func_name not in self._no_funcs:
                self._helper_funcs.append(func_str)

        if 'func' in func_info:

            # extract the provided callable
            func = func_info['func']

        else:

            # extract the provided function definition
            func_str = func_info['def']

            # make _imports available to function
            for imp in self._imports:
                try:
                    exec(imp, globals())
                except SyntaxError:
                    pass

            # evaluate the function string to receive a callable
            exec(func_str, globals())
            func = globals().pop(func_name)

        return {'func': func, 'call': func_name}

    def add_var_update(self, lhs: ComputeVar, rhs: str, lhs_idx: Optional[str] = None, rhs_shape: Optional[tuple] = ()):

        lhs = lhs.name
        if lhs_idx:
            idx, _ = self.create_index_str(lhs_idx, apply=True)
            lhs = f"{lhs}{idx}"
        self.add_code_line(f"{lhs} = {rhs}")

    def create_index_str(self, idx: Union[str, int, tuple], separator: str = ',', apply: bool = True,
                         **kwargs) -> Tuple[str, dict]:

        # preprocess idx
        if type(idx) is str and separator in idx:
            idx = tuple(idx.split(separator))

        # case: multiple indices
        if type(idx) is tuple:
            idx = list(idx)
            for i in range(len(idx)):
                idx[i] = self._process_idx(idx[i], **kwargs)
            idx = tuple([f"{i}" for i in idx])
            idx_str = f"{self._idx_left}{separator.join(idx)}{self._idx_right}" if apply else separator.join(idx)
            return idx_str, dict()

        # case: single index
        idx = self._process_idx(idx, **kwargs)
        return f"{self._idx_left}{idx}{self._idx_right}" if apply else idx, dict()

    def get_fname(self, f: str) -> tuple:

        f_split = f.split('.')
        if len(f_split) > 2:
            raise ValueError(f'File name {f} has wrong format. Only one `.` can be used to separate file name from '
                             f'file ending.')
        if len(f_split) == 2:
            *path, file = f_split[0].split('/')
            return '/'.join(path), file, f_split[1]
        else:
            *path, file = f.split('/')
            return '/'.join(path), file

    def generate_func_head(self, func_name: str, state_var: str = None, func_args: list = None):

        imports = self._imports
        helper_funcs = self._helper_funcs
        if func_args:
            func_args = [arg.name for arg in func_args]
        else:
            func_args = []
        state_vars = ['t', state_var]
        _, indices = np.unique(func_args, return_index=True)
        func_args = state_vars + [func_args[idx] for idx in np.sort(indices)]

        if imports:

            # add _imports at beginning of file
            for imp in imports:
                self.add_code_line(imp)
            self.add_linebreak()

        if helper_funcs:

            # add definitions of helper functions after the _imports
            for func in helper_funcs:
                self.add_code_line(func)
            self.add_linebreak()

        # add function header
        self.add_linebreak()
        self._add_func_call(name=func_name, args=func_args)
        self.add_indent()

        return func_args

    def generate_func_tail(self, rhs_var: str = None):

        self.add_code_line(f"return {rhs_var}")
        self.remove_indent()

    def generate_func(self, func_name: str, to_file: bool = True, **kwargs):

        # generate the current function string via the code generator
        func_str = self.generate()

        if to_file:

            # save rhs function to file
            file = f'{self._fdir}/{self._fname}' if self._fdir else self._fname
            with open(f'{file}{self._fend}', 'w') as f:
                f.writelines(func_str)
                f.close()

            # import function from file
            exec(f"from {self._fname} import {func_name}", globals())

        else:

            # just execute the function string, without writing it to file
            exec(func_str, globals())

        rhs_eval = globals().pop(func_name)

        # apply function decorator
        decorator = kwargs.pop('decorator', None)
        if decorator:
            decorator_kwargs = kwargs.pop('decorator_kwargs', dict())
            rhs_eval = decorator(rhs_eval, **decorator_kwargs)

        return rhs_eval

    def run(self, func: Callable, func_args: tuple, T: float, dt: float, dts: float, outputs: dict,
            solver: str, **kwargs) -> dict:

        # initial values
        t0 = func_args[0]
        times = np.arange(0.0, T, dts) if dts else np.arange(0.0, T, dt)
        y0 = func_args[1]

        # perform simulation
        results = self._solve(solver=solver, func=func, args=func_args[2:], T=T, dt=dt, dts=dts, y0=y0, t0=t0,
                              times=times, **kwargs)

        # reduce state recordings to requested state variables
        for key, idx in outputs.items():
            if type(idx) is tuple and idx[1] - idx[0] == 1:
                idx = (idx[0],)
            elif type(idx) is int:
                idx = (idx,)
            outputs[key] = results[:, idx] if len(idx) == 1 else results[:, idx[0]:idx[1]]
        outputs['time'] = times

        return outputs

    def clear(self):

        # clear code generator
        super().clear()

        # remove files and directories that have been created during simulation process
        if self._fdir:
            rmtree(self._fdir)
        else:
            try:
                os.remove(f"{self._fname}{self._fend}")
            except FileNotFoundError:
                pass

        # delete loaded modules from the system
        if self._fname in sys.modules:
            del sys.modules[self._fname]

    @staticmethod
    def register_vars(variables: list):
        pass

    @staticmethod
    def finalize_idx_str(var: ComputeVar, idx: str):
        return f"{var.name}{idx}"

    @staticmethod
    def expr_to_str(expr: str, args: tuple):
        return expr

    def _get_func_info(self, name: str, **kwargs):
        return self._funcs[name]

    def _process_idx(self, idx: Union[Tuple[int, int], int, str, ComputeVar], **kwargs) -> str:
        if type(idx) is ComputeVar:
            idx.set_value(idx.value + self._start_idx)
            return idx.name
        if type(idx) is tuple:
            return f"{idx[0] + self._start_idx}:{idx[1]}"
        if type(idx) is int:
            return f"{idx + self._start_idx}"
        try:
            return self._process_idx(int(idx), **kwargs)
        except (TypeError, ValueError):
            return idx

    def _solve(self, solver: str, func: Callable, args: tuple, T: float, dt: float, dts: float, y0: np.ndarray,
               t0: np.ndarray, times: np.ndarray, **kwargs) -> np.ndarray:

        # perform integration via scipy solver (mostly Runge-Kutta methods)
        if solver == 'euler':

            # solve ivp via forward euler method (fixed integration step-size)
            results = self._solve_euler(func, args, T, dt, dts, y0, t0)

        else:

            # solve ivp via scipy methods (solvers of various orders with adaptive step-size)
            from scipy.integrate import solve_ivp
            kwargs['t_eval'] = times

            # call scipy solver
            results = solve_ivp(fun=func, t_span=(t0, T), y0=y0, first_step=dt, args=args, **kwargs)
            results = results['y'].T

        return results

    def _add_func_call(self, name: str, args: Iterable):
        self.add_code_line(f"def {name}({','.join(args)}):")

    @staticmethod
    def _solve_euler(func: Callable, args: tuple, T: float, dt: float, dts: float, y: np.ndarray, t0: np.ndarray):

        # preparations for fixed step-size integration
        idx = 0
        steps = int(np.round(T / dt))
        store_steps = int(np.round(T / dts))
        store_step = int(np.round(dts / dt))
        state_rec = np.zeros((store_steps, y.shape[0]) if y.shape else (store_steps, 1), dtype=y.dtype)

        # solve ivp for forward Euler method
        for step in range(t0, steps):
            if step % store_step == 0:
                state_rec[idx, :] = y
                idx += 1
            rhs = func(step, y, *args)
            y += dt * rhs

        return state_rec



"""Test suite for accurate behavior of models provided under pyrates.model_templates"""

# external _imports
from typing import Union
import numpy as np
import pytest

# pyrates internal _imports
from pyrates import integrate

# meta infos
__author__ = "Richard Gast"
__status__ = "Development"


###########
# Utility #
###########

# define backends for which to run the tests
backends = ['default', 'torch', 'tensorflow']
vectorization = [False, True, True]
backend_kwargs = [{}, {}, {}]
complex_compat = [False, True, False]

# define test accuracy
accuracy = 1e-4


def setup_module():
    print("\n")
    print("================================")
    print("| Test Suite 3 : SimNet Module |")
    print("================================")


def nmrse(x: np.ndarray,
          y: np.ndarray
          ) -> Union[float, np.ndarray]:
    """Calculates the normalized root mean squared error of two vectors of equal length.
    Parameters
    ----------
    x,y
        Arrays to calculate the nmrse between.
    Returns
    -------
    float
        Normalized root mean squared error.
    """

    max_val = np.max((np.max(x, axis=0), np.max(y, axis=0)))
    min_val = np.min((np.min(x, axis=0), np.min(y, axis=0)))

    diff = x - y

    return np.sqrt(np.sum(diff ** 2, axis=0)) / (max_val - min_val)


#########
# Tests #
#########


def test_3_1_jansenrit():
    """Testing accuracy of Jansen-Rit model implementations:
    """

    T = 1.5
    dt = 1e-4
    dts = 1e-2

    for b, v, kwargs in zip(backends, vectorization, backend_kwargs):

        # compare single operator JRC implementation with multi-node JRC implementation
        ###############################################################################

        # single operator JRC
        r1 = integrate("model_templates.neural_mass_models.jansenrit.JRC2", simulation_time=T,
                       outputs={'EIN': 'jrc/jrc_op/V_ein'}, backend=b, step_size=dt, solver='scipy',
                       sampling_step_size=dts, clear=True, file_name='jrc1', vectorize=v, **kwargs)

        # multi-node JRC
        r2 = integrate("model_templates.neural_mass_models.jansenrit.JRC", simulation_time=T,
                       outputs={'EIN': 'ein/rpo_e/V'}, backend=b, step_size=dt, solver='scipy',
                       sampling_step_size=dts, clear=True, file_name='jrc2', vectorize=v, **kwargs)

        assert np.mean(r1.values.flatten() - r2.values.flatten()) == pytest.approx(0., rel=accuracy, abs=accuracy)


def test_3_2_qif_theta():
    """Testing accuracy of mean-field representation of QIF population and theta neuron model.
    """

    T = 80.0
    dt = 1e-3
    dts = 1e-2

    in_start = int(np.round(30.0/dt))
    in_dur = int(np.round(20.0/dt))
    inp = np.zeros((int(np.round(T/dt)),))
    inp[in_start:in_start+in_dur] = 5.0

    for b, kwargs, c in zip(backends, backend_kwargs, complex_compat):

        # compare qif population dynamics with and without plasticity
        #############################################################

        # basic qif population
        r1 = integrate("model_templates.neural_mass_models.qif.qif", simulation_time=T, sampling_step_size=dts,
                       inputs={"p/qif_op/I_ext": inp}, outputs={"r": "p/qif_op/r"}, method='RK45', backend=b,
                       solver='scipy', step_size=dt, clear=True, file_name='m1', vectorize=False, **kwargs)

        # qif population with spike-frequency adaptation
        r2 = integrate("model_templates.neural_mass_models.qif.qif_sfa", simulation_time=T, sampling_step_size=dts,
                       inputs={"p/qif_sfa_op/I_ext": inp}, outputs={"r": "p/qif_sfa_op/r"}, method='RK45', backend=b,
                       solver='scipy', step_size=dt, clear=True, file_name='m2', vectorize=False, **kwargs)

        # qif population with synaptic depression
        r3 = integrate("model_templates.neural_mass_models.qif.qif_sd", simulation_time=T, sampling_step_size=dts,
                       inputs={"p/qif_op/I_ext": inp}, outputs={"r": "p/qif_op/r"}, method='RK45', backend=b,
                       solver='scipy', step_size=dt, clear=True, file_name='m3', vectorize=False, **kwargs)

        # test firing rate relationships at pre-defined times
        time = 49.0
        idx = np.argmin(np.abs(time - r1.index))
        assert r1.iloc[idx, 0] > r2.iloc[idx, 0] > r3.iloc[idx, 0]

        # compare qif population dynamics with theta neuron dynamics
        ############################################################

        if c:

            # basic theta neuron population
            r4 = integrate("model_templates.neural_mass_models.theta.kmo_theta", simulation_time=T, sampling_step_size=dts,
                           inputs={"p/theta_op/I_ext": inp}, outputs={"z": "p/theta_op/z"}, backend=b,
                           step_size=dt, clear=True, file_name='m1', vectorize=False, **kwargs)

            # translate complex variable z into firing rate r
            z = r4["z"].values
            z_conj = np.conjugate(z)
            w = (1-z_conj)/(1+z_conj)
            r = np.real(w)/np.pi

            # compare firing rate dynamics between theta and qif neuron
            assert np.mean(r - r1["r"].values) == pytest.approx(0., rel=1e-2, abs=1e-2)


def test_3_3_wilson_cowan():
    """Test accuracy of wilson cowan neural mass model implementation.
    """

    # assess correct response of model to input
    ###########################################

    T = 120.0
    dt = 5e-4
    dts = 1e-2

    in_start = int(np.round(50.0 / dt))
    in_dur = int(np.round(30.0 / dt))
    inp = np.zeros((int(np.round(T / dt)),))
    inp[in_start:in_start + in_dur] = 1.0

    for b, v, kwargs in zip(backends, vectorization, backend_kwargs):

        # standard wilson-cowan model
        r1 = integrate("model_templates.neural_mass_models.wilsoncowan.WC", simulation_time=T,
                       sampling_step_size=dts, inputs={"e/se_op/r_ext": inp}, outputs={"R_e": "e/rate_op/r"},
                       backend=b, solver='scipy', step_size=dt, clear=True, file_name='wc1', vectorize=v, **kwargs)

        # wilson-cowan model with synaptic short-term plasticity
        r2 = integrate("model_templates.neural_mass_models.wilsoncowan.WC_stp", simulation_time=T,
                       sampling_step_size=dts, inputs={"e/se_op/r_ext": inp}, outputs={"R_e": "e/rate_op/r"},
                       backend=b, solver='scipy', step_size=dt, clear=True, file_name='wc2', vectorize=v, **kwargs)

        # test firing rate relationships at pre-defined times
        times = [75.0, 119.0]
        indices = [np.argmin(np.abs(t - r2.index)) for t in times]
        for idx in indices:
            assert r2.iloc[idx, 0] > r1.iloc[idx, 0]


def test_3_4_kuramoto():
    """Tests accurate behavior of kuramoto oscillator model.
    """

    T = 6.0
    dt = 1e-4
    dts = 1e-2

    in_start = int(np.round(0.0 / dt))
    in_dur = int(np.round(0.1 / dt))
    inp = np.zeros((int(np.round(T / dt)),))
    inp[in_start:in_start + in_dur] = 1.0

    for b, v, kwargs, c in zip(backends, vectorization, backend_kwargs, complex_compat):

        # assess correct response of single base oscillator
        ###################################################

        # perform simulation
        r1 = integrate("model_templates.oscillators.kuramoto.kmo", simulation_time=T, sampling_step_size=dts,
                       outputs={"theta": "p/phase_op/theta"}, backend=b, solver='scipy', step_size=dt, clear=True,
                       method='RK45', file_name='km1', vectorize=False, **kwargs)

        # test linear oscillator properties
        omega = 10.0
        results = np.sin(r1.values[:, 0]*2*np.pi)
        target = np.sin(omega*2.0*np.pi*r1.index.values)
        assert results - target == pytest.approx(0., rel=1e-2, abs=1e-2)

        # assess correct response of two coupled oscillators
        ####################################################

        # perform simulation
        r2 = integrate("model_templates.oscillators.kuramoto.kmo_2coupled", simulation_time=T, sampling_step_size=dts,
                       outputs={"theta1": "p1/phase_op/theta", "theta2": "p2/phase_op/theta"}, backend=b, solver='scipy',
                       inputs={"p1/phase_op/s_ext": inp}, step_size=dt, clear=True, file_name='km2', vectorize=v,
                       method='RK45', **kwargs)

        # test whether the oscillators expressed de-phasing
        init = int(in_dur*dt/dts)
        diff_init = r2['theta1'].iloc[init] - r2['theta2'].iloc[init]
        diff_end = r2['theta1'].iloc[-1] - r2['theta2'].iloc[-1]
        assert diff_end > diff_init

        # assess correct response of kuramoto order parameter mean-field model
        ######################################################################

        # backends that can't handle complex variables
        if c:

            # define oscillatory input
            t = np.linspace(0, T, num=int(np.round(T/dt)))
            inp = 2.0/(1.0 + np.exp(np.sin(2.0*np.pi*t*1.0)))

            # perform simulation
            r3 = integrate("model_templates.oscillators.kuramoto.kmo_mf_2coupled", simulation_time=T,
                           sampling_step_size=dts, outputs={"z1": "p1/kmo_op/z", "z2": "p2/kmo_op/z"}, backend=b,
                           step_size=dt, solver='scipy', clear=True, file_name='km3', vectorize=v, **kwargs)

            # perform simulation
            r4 = integrate("model_templates.oscillators.kuramoto.kmo_mf_2coupled", simulation_time=T,
                           sampling_step_size=dts, outputs={"z1": "p1/kmo_op/z", "z2": "p2/kmo_op/z"}, backend=b,
                           inputs={"p1/kmo_op/s_ext": inp}, step_size=dt,
                           solver='scipy', clear=True, file_name='km4', vectorize=v, **kwargs)

            # verify that the stimulated model shows stronger synchronization than un-stimulated model
            assert np.abs(r3.iloc[-1, 0]) < np.abs(r4.iloc[-1, 0])
            assert np.abs(r3.iloc[-1, 1]) < np.abs(r4.iloc[-1, 1])




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from typing import List

from pyrates.backend import PyRatesException
from pyrates.ir.node import VectorizedNodeIR, NodeIR
from pyrates.ir.operator_graph import OperatorGraph

__author__ = "Daniel Rose"
__status__ = "Development"


class EdgeIR(VectorizedNodeIR):

    __slots__ = VectorizedNodeIR.__slots__ + ["_inputs"]

    def __init__(self, label: str, operators: OperatorGraph = None, values: dict = None, template: str = None):

        if not operators:
            operators = {}
            values = {}

        super().__init__(label=label, operators=operators, values=values, template=template)

        self._inputs = None
        self.length = 1

    @property
    def inputs(self):
        """Detect input variables of edge. This also references the operator
         the variable belongs to.

         Note: As of `0.9.0` multiple source/input variables are allowed per edge."""

        inputs = self.get_inputs()

        if len(inputs) == 0:
            return None
        else:
            return inputs

    @property
    def n_inputs(self):
        """Computes number of input variables that need to be informed from outside the edge (meaning from some node).
        """
        inputs = self.get_inputs()
        return len(inputs)

    def get_inputs(self):
        """Find all input variables of operators that need to be mapped to source variables in a source node."""
        try:
            len(self._inputs)
        except TypeError:
            # find inputs
            # noinspection PyTypeChecker
            # in_op = [op for op, in_degree in self.op_graph.in_degree if in_degree == 0]  # type: List[str]
            inputs = dict()
            # alternative:
            for op, op_data in self.op_graph.nodes(data=True):
                for var, var_data in op_data["inputs"].items():
                    if len(var_data["sources"]) == 0:
                        key = f"{op}/{var}"
                        try:
                            inputs[var].append(key)
                        except KeyError:
                            inputs[var] = [key]

            self._inputs = inputs
        return self._inputs

    @property
    def output(self):
        """Detect output variable of edge, assuming only one output variable exists."""

        # try to find single output variable
        # noinspection PyTypeChecker
        out_op = [op for op, out_degree in self.op_graph.out_degree if out_degree == 0]  # type: List[str]

        # only one single output operator allowed
        if len(out_op) == 1:
            out_var = self[out_op[0]]['output']
            return f"{self.label}/{out_op[0]}/{out_var}"
        elif len(out_op) == 0:
            return None
        else:
            raise PyRatesException("Too many or too little output operators found. Exactly one output operator and "
                                   "associated output variable is required per edge.")

    def __hash__(self):
        raise NotImplementedError




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Contains Julia function definitions that may be used for PyRates model equations.
"""

# external _imports
import numpy as np

# meta infos
__author__ = "Richard Gast"
__status__ = "development"

# function definitions
######################

sigmoid_func = lambda x: 1./(1. + np.exp(-x))
sigmoid_def = """
function sigmoid(x)
    return 1.0/(1.0 + exp(-x))
end
"""

interp = """
function interp(x_new, x, y)
    idx = argmin(abs.(x.-x_new))
    if x[idx] > x_new
        i1, i2 = max(idx-1, 1), idx
    else
        i1, i2 = idx, min(idx+1, length(y))
    end
    return (y[i1] + y[i2])*0.5
end
"""

# dictionary for backend import
###############################

julia_funcs = {
    'max': {'call': 'maximum', 'func': np.maximum, 'imports': []},
    'min': {'call': 'mininum', 'func': np.minimum, 'imports': []},
    'round': {'call': 'round', 'func': np.round, 'imports': []},
    'sum': {'call': 'sum', 'func': np.sum, 'imports': []},
    'mean': {'call': 'mean', 'func': np.mean, 'imports': []},
    'matmul': {'call': '*', 'func': np.dot, 'imports': []},
    'matvec': {'call': '*', 'func': np.dot, 'imports': []},
    'roll': {'call': 'roll', 'func': np.roll, 'imports': []},
    'randn': {'call': 'normal', 'func': np.random.randn, 'imports': []},
    'tanh': {'call': 'tanh', 'func': np.tanh, 'imports': []},
    'sinh': {'call': 'sinh', 'func': np.sinh, 'imports': []},
    'cosh': {'call': 'cosh', 'func': np.cosh, 'imports': []},
    'arctan': {'call': 'atan', 'func': np.arctan, 'imports': []},
    'arcsin': {'call': 'asin', 'func': np.arcsin, 'imports': []},
    'arccos': {'call': 'acos', 'func': np.arccos, 'imports': []},
    'sin': {'call': 'sin', 'func': np.sin, 'imports': []},
    'cos': {'call': 'cos', 'func': np.cos, 'imports': []},
    'tan': {'call': 'tan', 'func': np.tan, 'imports': []},
    'exp': {'call': 'exp', 'func': np.exp, 'imports': []},
    'sigmoid': {'call': 'sigmoid', 'func': sigmoid_func, 'def': sigmoid_def, 'imports': []},
    'interp': {'call': 'interp', 'func': np.interp, 'def': interp, 'imports': []},
    'real': {'call': 'real', 'func': np.real, 'imports': []},
    'imag': {'call': 'imag', 'func': np.imag, 'imports': []},
    'conj': {'call': 'conj', 'func': np.conjugate, 'imports': []},
    'absv': {'call': 'abs', 'func': np.abs, 'imports': []},
}




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
from typing import Union, List, Type, Dict

from pyrates.ir.circuit import PyRatesException
from pyrates.frontend.template import _complete_template_path
from pyrates.frontend.template.abc import AbstractBaseTemplate
from pyrates.frontend.template.operator import OperatorTemplate


class OperatorGraphTemplate(AbstractBaseTemplate):

    def __init__(self, name: str, path: str, operators: Union[str, List[str], dict, List[AbstractBaseTemplate]],
                 description: str = "A node or an edge."):
        """For now: only allow single equation in operator template."""

        super().__init__(name, path, description)

        self.operators = {}  # dictionary with operator path as key and variations to the template as values
        self._op_map = {}
        if isinstance(operators, str):
            operator_template = self._load_operator_template(operators)
            self.operators[operator_template] = {}  # single operator path with no variations
            self._op_map[operator_template.name] = operator_template
        elif isinstance(operators, list):
            for op in operators:
                try:
                    operator_template = self._load_operator_template(op)
                except TypeError:
                    operator_template = op
                self.operators[operator_template] = {}  # multiple operator paths with no variations
                self._op_map[operator_template.name] = operator_template
        elif isinstance(operators, dict):
            for op, variations in operators.items():
                try:
                    operator_template = self._load_operator_template(op)
                except TypeError:
                    operator_template = op
                self.operators[operator_template] = variations
                self._op_map[operator_template.name] = operator_template
        # for op, variations in operators.items():
        #     if "." not in op:
        #         op = f"{path.split('.')[:-1]}.{op}"
        #     self.operators[op] = variations

    def update_template(self, name: str = None, path: str = None, operators: Union[str, List[str], dict] = None,
                        description: str = None):
        """Update all entries of a base edge template to a more specific template."""

        if not name:
            name = self.name

        if not path:
            path = self.path

        if not description:
            description = self.__doc__

        if operators:
            operators = _update_operators(self.operators, operators)
        else:
            operators = self.operators

        return self.__class__(name=name, path=path, operators=operators, description=description)

    def update_var(self, op: str, var: str, val: Union[float, int]):

        op_template = self.get_op(op)
        self.operators[op_template][var] = val

    def get_op(self, op: str):
        return self._op_map[op]

    def apply(self, values: dict = None, label: str = None, **kwargs):
        """ Apply template to gain a node or edge intermediate representation.

        Parameters
        ----------
        values
            dictionary with operator/variable as keys and values to update these variables as items.
        label
            Label of the resulting intermediate representation. If `None`, the name of the template will be used.

        Returns
        -------

        """
        if not label:
            label = self.name

        value_updates = {}
        if values:
            for key, value in values.items():
                op_name, var_name = key.split("/")
                if op_name not in value_updates:
                    value_updates[op_name] = {}
                value_updates[op_name][var_name] = value

        operators = {}
        all_values = {}  # # type: Dict[OperatorIR, Dict]

        for template, variations in self.operators.items():
            values_to_update = variations

            if values_to_update is None:
                values_to_update = {}
            # if a value for this particular variation has been passed, overwrite the previous value
            if template.name in value_updates:
                values_to_update.update(value_updates.pop(template.name))
            # apply operator template to get OperatorIR and associated default values and label
            operator, op_values, key = template.apply(return_key=True, values=values_to_update)
            operators[key] = operator
            all_values[key] = op_values

        # fail gracefully, if any variables remain in `values` which means, that there is some typo
        if value_updates:
            raise PyRatesException(
                "Found value updates that did not fit any operator by name. This may be due to a "
                "typo in specifying the operator or variable to update. Remaining variables:"
                f"{value_updates}")
        return self.target_ir(label, operators, values=all_values, template=self.path, **kwargs)

    @staticmethod
    def target_ir(*args, **kwargs):
        raise NotImplementedError

    def _load_operator_template(self, path: str) -> OperatorTemplate:
        """Load an operator template based on a path"""
        path = _complete_template_path(path, self.path)
        return OperatorTemplate.from_yaml(path)


def _update_operators(base_operators: dict, updates: Union[str, List[str], dict]):
    """Update operators of a given template.

    Parameters:
    -----------

    base_operators:
        Reference to one or more operators in the base class.
    updates:
        Reference to one ore more operators in the child class
        - string refers to path or name of single operator
        - list refers to multiple operators of the same class
        - dict contains operator path or name as key
    """
    updated = {}

    if isinstance(updates, str):
        updated[updates] = {}  # single operator path with no variations
    elif isinstance(updates, list):
        for path in updates:
            updated[path] = {}  # multiple operator paths with no variations
    elif isinstance(updates, dict):
        for path, variations in updates.items():
            updated[path] = variations
        # dictionary with operator path as key and variations as sub-dictionary
    else:
        raise TypeError("Unable to interpret type of operator updates. Must be a single string,"
                        "list of strings or dictionary.")
    # # Check somewhere, if child operators have same input/output as base operators?
    #
    return updated







"""Interfaces for loading and saving (dumping) templates and circuits from/to file."""
from typing import Optional

FILEIOMODES = ["pickle", "yaml"]


def save(_obj: object, filename: str, filetype: str, template_name: Optional[str] = None, **kwargs):
    """Save an object to file.

    Parameters
        ----------
        _obj
            The object to dump to file. It must either be pickleable or have a `to_dict` method for 'yaml' mode.
        filename
            Path to file (absolute or relative).
        filetype
            Chooses which loader to use to load the file. Allowed types: pickle, yaml
        template_name
            This argument is needed for filetype 'yaml'. The object is saved under that name within the file.

        Returns
        -------
        None
    """

    if filetype == "pickle":
        # make sure the directory exists
        from pyrates.utility import create_directory
        create_directory(filename)

        from pyrates.frontend.fileio import pickle
        pickle.dump(_obj, filename, **kwargs)
        print(f"{_obj} successfully pickled to {filename}")

    elif filetype == "yaml":
        try:
            _dict = _obj.to_dict()  # type: dict
        except AttributeError:
            try:
                _dict = dict(dict(nodes=_obj.nodes, edges=_obj.edges, base='CircuitTemplate'))
            except AttributeError:
                raise AttributeError(f"The object {_obj} does not have a `to_dict` method and is no instance of "
                                     f"`CircuitTemplate`. Please choose a different target filetype than 'yaml'.")

        # make sure the directory exists
        from pyrates.utility import create_directory
        create_directory(filename)

        from pyrates.frontend.fileio import yaml
        yaml.dump(_dict, filename, template_name, **kwargs)
        print(f"{_obj} successfully dumped to {filename} in YAML format.")

    else:

        raise ValueError(f"Unknown file format to save to. Allowed modes: {FILEIOMODES}")



from setuptools import setup, find_packages

PYRATES_TEAM = "Richard Gast, Daniel Rose"

INSTALL_REQUIREMENTS = ['numpy',
                        'networkx',
                        'pandas',
                        'ruamel.yaml',
                        'sympy',
                        'scipy'
                        ]

CLASSIFIERS = ["Programming Language :: Python :: 3",
               "Programming Language :: Python :: 3.6",
               "Programming Language :: Python :: 3.7",
               "Programming Language :: Python :: 3.8",
               "Programming Language :: Python :: 3.9",
               "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
               "Operating System :: OS Independent",
               "Development Status :: 3 - Alpha",
               "Intended Audience :: Developers",
               "Intended Audience :: Science/Research",
               "Topic :: Scientific/Engineering",
               ]

EXTRAS = {"backends": ["tensorflow>=2.0", "torch"],
          "dev": ["pytest", "bump2version"]}

EXTRAS["all"] = [item for sublist in EXTRAS.values() for item in sublist]
EXTRAS["tests"] = [item for key, sublist in EXTRAS.items() for item in sublist if key in ("backends", "dev")]

with open("VERSION", "r", encoding="utf8") as fh:
    VERSION = fh.read().strip()

with open("README.md", "r", encoding="utf8") as fh:
    DESCRIPTION = fh.read()

setup(name='pyrates',
      version=VERSION,
      description='Dynamical Systems Modeling Framework',
      long_description=DESCRIPTION,
      long_description_content_type='text/markdown',
      author=PYRATES_TEAM,
      author_email='richard.gast@northwestern.edu',
      license='GPL v3',
      packages=find_packages(),
      zip_safe=False,
      python_requires='>=3.6',
      install_requires=INSTALL_REQUIREMENTS,
      extras_require=EXTRAS,
      classifiers=CLASSIFIERS,
      include_package_data=True  # include additional non-python files specified in MANIFEST.in
      )



# """Test suite for cluster computation functionality.
# """
# import pytest
#
# # meta infos
# __author__ = "Richard Gast, Daniel Rose"
# __status__ = "Development"
#
# ###########
# # Utility #
# ###########
#
#
# def setup_module():
#     print("\n")
#     print("===================================")
#     print("| Test Suite: Cluster computation |")
#     print("===================================")
#
#
# #########
# # Tests #
# #########
#
#
# def test_1_cluster_compute(tmp_path):
#     import platform
#     import pandas as pd
#     import numpy as np
#
#     from pyrates.utility.grid_search import ClusterGridSearch
#
#     nodes = [platform.node()]
#
#     # Simulation specs
#     circuit_template = "model_templates.jansen_rit.simple_jansenrit.JRC"
#     dt = 1e-3
#     sim_time = 1.0
#     param_map = {'u': {'vars': ['RPO_e_pc/u'],
#                        'nodes': ['PC']}}
#
#     # Create param_grid file
#     u = np.linspace(100, 400, 10)
#     param_grid = {'u': u}
#
#     output = {'r': 'PC/PRO/m_out'}
#     cgs = ClusterGridSearch(nodes=nodes, compute_dir=tmp_path)
#     res_file = cgs.run(
#         circuit_template=circuit_template,
#         param_grid=param_grid,
#         param_map=param_map,
#         simulation_time=sim_time,
#         dt=dt,
#         permute_grid=False,
#         sampling_step_size=dt,
#         inputs={},
#         outputs=output,
#         chunk_size=len(u),
#         add_template_info=False,
#         verbose=True,
#         gs_kwargs={
#             "init_kwargs": {
#                 'backend': 'numpy',
#                 'solver': 'euler'
#             },
#             "njit": True,
#             "parallel": False
#         },
#         worker_kwargs={"test": "test"}
#     )
#
#     try:
#         results = pd.read_hdf(res_file, key=f'Results/results')
#     except KeyError:
#         assert False
#     assert not results.empty
#
